// Copyright 2014 BVLC and contributors.

#include <vector>

#include "caffe/layer.hpp"
#include "caffe/vision_layers.hpp"
#include "caffe/util/im2col.hpp"
#include "caffe/filler.hpp"
#include "caffe/util/math_functions.hpp"

namespace caffe {

template <typename Dtype>
void ConvolutionOrthLayer<Dtype>::SetUp(const vector<Blob<Dtype>*>& bottom,
      vector<Blob<Dtype>*>* top) {
  Layer<Dtype>::SetUp(bottom, top);
  kernel_size_ = this->layer_param_.convolution_param().kernel_size();
  stride_ = this->layer_param_.convolution_param().stride();
  group_ = this->layer_param_.convolution_param().group();
  pad_ = this->layer_param_.convolution_param().pad();
  num_ = bottom[0]->num();
  channels_ = bottom[0]->channels();
  height_ = bottom[0]->height();
  width_ = bottom[0]->width();
  num_output_ = this->layer_param_.convolution_param().num_output();
  CHECK_GT(num_output_, 0);
  CHECK_EQ(channels_ % group_, 0);
  // The im2col result buffer would only hold one image at a time to avoid
  // overly large memory usage.
  int height_out = (height_ + 2 * pad_ - kernel_size_) / stride_ + 1;
  int width_out = (width_ + 2 * pad_ - kernel_size_) / stride_ + 1;
  col_buffer_.Reshape(
      1, channels_ * kernel_size_ * kernel_size_, height_out, width_out);
  // Set the parameters
  CHECK_EQ(num_output_ % group_, 0)
      << "Number of output should be multiples of group.";
  bias_term_ = this->layer_param_.convolution_param().bias_term();
  // Figure out the dimensions for individual gemms.
  M_ = num_output_ / group_;
  K_ = channels_ * kernel_size_ * kernel_size_ / group_;
  N_ = height_out * width_out;
  (*top)[0]->Reshape(bottom[0]->num(), num_output_, height_out, width_out);
  // Check if we need to set up the weights
  if (this->blobs_.size() > 0) {
    LOG(INFO) << "Skipping parameter initialization";
  } else {
    if (bias_term_) {
      this->blobs_.resize(2);
    } else {
      this->blobs_.resize(1);
    }
    // Intialize the weight
    this->blobs_[0].reset(new Blob<Dtype>(
        num_output_, channels_ / group_, kernel_size_, kernel_size_));
    // fill the weights
    shared_ptr<Filler<Dtype> > weight_filler(GetFiller<Dtype>(
        this->layer_param_.convolution_param().weight_filler()));
    weight_filler->Fill(this->blobs_[0].get());
    // If necessary, intiialize and fill the bias term
    if (bias_term_) {
      this->blobs_[1].reset(new Blob<Dtype>(1, 1, 1, num_output_));
      shared_ptr<Filler<Dtype> > bias_filler(GetFiller<Dtype>(
          this->layer_param_.convolution_param().bias_filler()));
      bias_filler->Fill(this->blobs_[1].get());
    }
  }
  // Set up the bias filler
  if (bias_term_) {
    bias_multiplier_.reset(new SyncedMemory(N_ * sizeof(Dtype)));
    Dtype* bias_multiplier_data =
        reinterpret_cast<Dtype*>(bias_multiplier_->mutable_cpu_data());
    for (int i = 0; i < N_; ++i) {
        bias_multiplier_data[i] = 1.;
    }
  }
  // Setting up variables needed for orthogonalization 
  LOG(INFO) << "Convolutional orth layer: M_=" << M_ << ", K_=" << K_;
  iter_ = 0;
  orth_step_ = this->layer_param_.orth_param().step();
  orth_method_ = this->layer_param_.orth_param().orth_method();
  max_num_iter_ = this->layer_param_.orth_param().max_num_iter();
  eps_ = this->layer_param_.orth_param().eps();
  esmaeili_coeff_ = this->layer_param_.orth_param().esmaeili_coeff();
  min_norm_ = this->layer_param_.orth_param().min_norm();
  max_norm_ = this->layer_param_.orth_param().max_norm();
  target_norm_ = this->layer_param_.orth_param().target_norm();
  col_norm_ = this->layer_param_.orth_param().col_norm();
  orth_before_iter_ = this->layer_param_.orth_param().orth_before_iter();
  this->gram_.Reshape(1, 1, M_, M_);
  this->kk_.Reshape(1, 1, M_, M_);
  this->id_.Reshape(1, 1, M_, M_);
  this->ak_.Reshape(1, 1, M_, K_);
  Dtype* id_data = this->id_.mutable_cpu_data();
  for (int i=0; i < M_; i++) {
    for (int j=0; j < M_; j++) {
      if (i==j)
        id_data[i * M_ + j] = 1.;
      else
        id_data[i * M_ + j] = 0.;
    }
  }
}


template <typename Dtype>
Dtype ConvolutionOrthLayer<Dtype>::Forward_cpu(const vector<Blob<Dtype>*>& bottom,
      vector<Blob<Dtype>*>* top) {
  const Dtype* bottom_data = bottom[0]->cpu_data();
  Dtype* top_data = (*top)[0]->mutable_cpu_data();
  Dtype* col_data = col_buffer_.mutable_cpu_data();
  Dtype* weight = this->blobs_[0]->mutable_cpu_data();
  int weight_offset = M_ * K_;
  int col_offset = K_ * N_;
  int top_offset = M_ * N_;
  
  // Orthogonalization  
  if (Caffe::phase() == Caffe::TRAIN && orth_step_ > 0) {
    if (!(iter_ % orth_step_) && (orth_before_iter_ == 0 || iter_ < orth_before_iter_)) {
//       LOG(INFO) << "Orthogonalizing, iter=" << iter_;
      switch (orth_method_) {
        case OrthParameter_OrthMethod_ESMAEILI:
        { 
          Dtype* gram = this->gram_.mutable_cpu_data();
          Dtype* kk = this->kk_.mutable_cpu_data();
          Dtype* ak = this->ak_.mutable_cpu_data();
          const Dtype* id = this->id_.cpu_data();
          Dtype error;
          caffe_scal(weight_offset * group_, Dtype(1) / col_norm_, weight);
          
          for (int g = 0; g < group_; ++g) {
//           LOG(INFO) << "ESMAEILI";                 
            for (int ni=0; ni<max_num_iter_; ni++) {
              caffe_cpu_gemm<Dtype>(CblasNoTrans, CblasTrans, M_, M_, K_, (Dtype)1.,
                  weight, weight, (Dtype)0., gram);
              caffe_cpu_gemm<Dtype>(CblasNoTrans, CblasNoTrans, M_, M_, M_, esmaeili_coeff_,
                  gram, gram, (Dtype)0., kk);
              caffe_axpy<Dtype>(M_*M_, -(1. + esmaeili_coeff_), gram, kk);
              caffe_axpy<Dtype>(M_*M_, 2., id, kk);            
              caffe_axpy<Dtype>(M_*M_, -1., id, gram);
              error = caffe_cpu_norm2(M_*M_, gram);
  //             LOG(INFO) << "Iter " << ni+1 << "  ||Gram - id||=" << error;
//               if (error > 1e5)
              if (error < eps_)
                ni = max_num_iter_;
              else {
                LOG(INFO) << "Iter " << iter_ << "." << ni <<"  ||Gram - id||=" << error;
                caffe_cpu_gemm<Dtype>(CblasNoTrans, CblasNoTrans, M_, K_, M_, (Dtype)1.,
                  kk, weight, (Dtype)0., ak);
                caffe_copy<Dtype>(M_*K_, ak, weight);
              }
            }
            weight += weight_offset;
          }
          weight = this->blobs_[0]->mutable_cpu_data();
          caffe_scal(weight_offset * group_, col_norm_, weight);
          
          break;
        }
        case OrthParameter_OrthMethod_NORM_L2:
        {
          normalize_weights(min_norm_, max_norm_, target_norm_);
          break;
        }
        case OrthParameter_OrthMethod_NONE:
//           LOG(INFO) << "NONE";
          break;
        default:
          LOG(FATAL) << "Unknown orthogonalization method";
          break;  
      }
    }      
    iter_++;
  }
  
  // actually performing forward pass
  
  // reset weight pointer
  weight = this->blobs_[0]->mutable_cpu_data();
  
  for (int n = 0; n < num_; ++n) {
    // First, im2col
    im2col_cpu(bottom_data + bottom[0]->offset(n), channels_, height_,
                      width_, kernel_size_, pad_, stride_, col_data);
    // Second, innerproduct with groups
    for (int g = 0; g < group_; ++g) {
      caffe_cpu_gemm<Dtype>(CblasNoTrans, CblasNoTrans, M_, N_, K_,
        (Dtype)1., weight + weight_offset * g, col_data + col_offset * g,
        (Dtype)0., top_data + (*top)[0]->offset(n) + top_offset * g);
    }
    // third, add bias
    if (bias_term_) {
      caffe_cpu_gemm<Dtype>(CblasNoTrans, CblasNoTrans, num_output_,
          N_, 1, (Dtype)1., this->blobs_[1]->cpu_data(),
          reinterpret_cast<const Dtype*>(bias_multiplier_->cpu_data()),
          (Dtype)1., top_data + (*top)[0]->offset(n));
    }
  }
  
  return Dtype(0.);
}

template <typename Dtype>
void ConvolutionOrthLayer<Dtype>::Backward_cpu(const vector<Blob<Dtype>*>& top,
      const bool propagate_down, vector<Blob<Dtype>*>* bottom) {
  const Dtype* top_diff = top[0]->cpu_diff();
  const Dtype* weight = this->blobs_[0]->cpu_data();
  Dtype* weight_diff = this->blobs_[0]->mutable_cpu_diff();
  const Dtype* bottom_data = (*bottom)[0]->cpu_data();
  Dtype* bottom_diff = (*bottom)[0]->mutable_cpu_diff();
  Dtype* col_data = col_buffer_.mutable_cpu_data();
  Dtype* col_diff = col_buffer_.mutable_cpu_diff();
  // bias gradient if necessary
  Dtype* bias_diff = NULL;

  if (bias_term_) {
    bias_diff = this->blobs_[1]->mutable_cpu_diff();
    memset(bias_diff, 0, sizeof(Dtype) * this->blobs_[1]->count());
    for (int n = 0; n < num_; ++n) {
      caffe_cpu_gemv<Dtype>(CblasNoTrans, num_output_, N_,
          1., top_diff + top[0]->offset(n),
          reinterpret_cast<const Dtype*>(bias_multiplier_->cpu_data()), 1.,
          bias_diff);
    }
  }

  int weight_offset = M_ * K_;
  int col_offset = K_ * N_;
  int top_offset = M_ * N_;
  memset(weight_diff, 0, sizeof(Dtype) * this->blobs_[0]->count());
  for (int n = 0; n < num_; ++n) {
    // since we saved memory in the forward pass by not storing all col data,
    // we will need to recompute them.
    im2col_cpu(bottom_data + (*bottom)[0]->offset(n), channels_, height_,
                      width_, kernel_size_, pad_, stride_, col_data);
    // gradient w.r.t. weight. Note that we will accumulate diffs.
    for (int g = 0; g < group_; ++g) {
      caffe_cpu_gemm<Dtype>(CblasNoTrans, CblasTrans, M_, K_, N_,
        (Dtype)1., top_diff + top[0]->offset(n) + top_offset * g,
        col_data + col_offset * g, (Dtype)1.,
        weight_diff + weight_offset * g);
    }
    // gradient w.r.t. bottom data, if necessary
    if (propagate_down) {
      for (int g = 0; g < group_; ++g) {
        caffe_cpu_gemm<Dtype>(CblasTrans, CblasNoTrans, K_, N_, M_,
          (Dtype)1., weight + weight_offset * g,
          top_diff + top[0]->offset(n) + top_offset * g,
          (Dtype)0., col_diff + col_offset * g);
      }
      // col2im back to the data
      col2im_cpu(col_diff, channels_, height_, width_, kernel_size_, pad_,
          stride_, bottom_diff + (*bottom)[0]->offset(n));
    }
  }
}

template <typename Dtype>
void ConvolutionOrthLayer<Dtype>::normalize_weights(Dtype mnorm) {
  Dtype *weight = 0;
  int M = this->blobs_[0]->channels() * this->blobs_[0]->width() * this->blobs_[0]->height();
  int N = this->blobs_[0]->num();
  int off = this->blobs_[0]->offset(1, 0, 0, 0);
  switch (Caffe::mode()) {
  case Caffe::CPU:
      // apply the constraint to each column
      weight = this->blobs_[0]->mutable_cpu_data();
      for (int i = 0; i < N; ++i) {
          // compute l2 norm
          Dtype nrm = caffe_cpu_norm2(M, weight, 1);
          if (nrm > mnorm) {
              // and scale
              caffe_scal(M, mnorm / (nrm + Dtype(1e-7)), weight, 1);
          }
          weight += off;
      }
      break;
  case Caffe::GPU:
      // apply the constraint to each column
      weight = this->blobs_[0]->mutable_gpu_data();
      for (int i = 0; i < N; ++i) {
          // compute l2 norm
          Dtype nrm = caffe_gpu_norm2(M, weight, 1);
          if (nrm > mnorm) {
              // and scale
              caffe_gpu_scal(M, mnorm / (nrm + Dtype(1e-7)), weight, 1);
          }
          weight += off;
      }
    break;
  default:
    LOG(FATAL) << "Unknown caffe mode.";
    break;
  }
}

template <typename Dtype>
void ConvolutionOrthLayer<Dtype>::normalize_weights(Dtype min_norm, Dtype max_norm, Dtype target_norm) {
  Dtype *weight = 0;
  int M = this->blobs_[0]->channels() * this->blobs_[0]->width() * this->blobs_[0]->height();
  int N = this->blobs_[0]->num();
  int off = this->blobs_[0]->offset(1, 0, 0, 0);
  switch (Caffe::mode()) {
  case Caffe::CPU:
      // apply the constraint to each column
      weight = this->blobs_[0]->mutable_cpu_data();
      for (int i = 0; i < N; ++i) {
          // compute l2 norm
          Dtype nrm = caffe_cpu_norm2(M, weight, 1);
          if (nrm < min_norm || nrm > max_norm) {
              // and scale
              caffe_scal(M, target_norm / (nrm + Dtype(1e-7)), weight, 1);
          }
          weight += off;
      }
      break;
  case Caffe::GPU:
      // apply the constraint to each column
      weight = this->blobs_[0]->mutable_gpu_data();
      for (int i = 0; i < N; ++i) {
          // compute l2 norm
          Dtype nrm = caffe_gpu_norm2(M, weight, 1);
          if (nrm > max_norm || nrm < min_norm) {
              // and scale
              caffe_gpu_scal(M, target_norm / (nrm + Dtype(1e-7)), weight, 1);
          }
          weight += off;
      }
    break;
  default:
    LOG(FATAL) << "Unknown caffe mode.";
    break;
  }
}

INSTANTIATE_CLASS(ConvolutionOrthLayer);

}  // namespace caffe
